/*
 * yasapi_config.c
 * Copyright (C) 2015 Peter Belkner <pbelkner@snafu.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <yasapi.h>
#if ! defined (YASAPI_CONFIG2) // {
#include <uxtheme.h>
#include <resources.h>

///////////////////////////////////////////////////////////////////////////////
typedef struct _PageVMT PageVMT;
typedef struct _PageTemplate PageTemplate;
typedef struct _Page Page;
typedef struct _Priv Priv;

///////////////////////////////////////////////////////////////////////////////
static void ConfigSet(Priv *pPriv, HWND hDlg);

///////////////////////////////////////////////////////////////////////////////
struct _PageVMT {
  const char *name;
  DLGPROC lpDialogFunc;
  void (*Set)(Page *pPage, Priv *pPriv);
  void (*Get)(Page *pPage, Priv *pPriv);
  void (*Reset)(Page *pPage, Priv *pPriv);
};

struct _PageTemplate {
  char *label;
  int idd;
  PageVMT *(*GetVMT)(void);
};

struct _Page {
  PageVMT *vmt;
  HWND hDlg;
};

struct _Priv {
  Player *pPlayer;
  HMODULE hModule;
  HWND hDlg;
  HWND hDlgPage;
  UINT uDevice;
  int nNumPages;
  int nPage;
  PageTemplate *pTemplates;
  Page aPages[0];
};

///////////////////////////////////////////////////////////////////////////////
enum {
  PAGE_DEVICES,
  PAGE_OPTIONS,
  PAGE_BUFFERS
};

static PageVMT *GetDevicesVMT(void);
static PageVMT *GetOptionsVMT(void);
static PageVMT *GetBuffersVMT(void);

static PageTemplate gaTemplates[]={
  { "Devices",IDD1_PAGE_DEVICES,GetDevicesVMT },
  { "Device Options",IDD1_PAGE_OPTIONS,GetOptionsVMT },
  { "Device Buffers",IDD1_PAGE_BUFFERS,GetBuffersVMT }
};

///////////////////////////////////////////////////////////////////////////////
static void SyncDevicePeriod(Priv *pPriv)
{
#if 0 // {
#if defined (YASAPI_DEBUG) // {
  int debug=pPriv->pPlayer->pPlugin->options.debug;
#endif // }
#endif // }
  HWND hDlgOptions=pPriv->aPages[PAGE_OPTIONS].hDlg;
  HWND hDlgBuffers=pPriv->aPages[PAGE_BUFFERS].hDlg;
  HWND hWndShareMode=GetDlgItem(hDlgOptions,IDC_COMBOBOX_SHAREMODE);
  HWND hWndPull=GetDlgItem(hDlgOptions,IDC_COMBOBOX_PULL);
  BOOL bShared=YASAPI_SHAREMODE_SHARE==SendMessageW(hWndShareMode,
      CB_GETCURSEL,0,0);
  BOOL bPull=SendMessageW(hWndPull,CB_GETCURSEL,0,0);
  BOOL bDevicePeriod=!(bShared&&bPull);

#if 0 // {
  DPRINTF(1,0,"hWndShareMode: %p, hWndPull: %p, bShared: %d, bPull: %d"
      ", bDevicePeriod: %d\n",
      hWndShareMode,hWndPull,bShared,bPull,bDevicePeriod);
#endif // }

  EnableWindow(GetDlgItem(hDlgBuffers,IDC_STATIC_LABEL_DEVICE_PERIOD),
      bDevicePeriod);
  EnableWindow(GetDlgItem(hDlgBuffers,IDC_COMBOBOX_DEVICE_PERIOD),
      bDevicePeriod);
  EnableWindow(GetDlgItem(hDlgBuffers,IDC_SLIDER_SHARE_SIZE),bDevicePeriod);
  EnableWindow(GetDlgItem(hDlgBuffers,IDC_STATIC_SHARE_SIZE),bDevicePeriod);
}

///////////////////////////////////////////////////////////////////////////////
static void DevicesFillList(HWND hDlg, Priv *pPriv, int idc)
{
  Player *pPlayer=pPriv->pPlayer;
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->pPlugin->options.debug;
#endif // }
  UINT cDevices=pPlayer->cDevices,uDevice;
  Channel *pChannel=pPlayer->pChannel;
  HWND hWndList;

  if (NULL==(hWndList=GetDlgItem(hDlg,idc))) {
    DMESSAGE(debug,"getting list");
    goto list;
  }

  for (uDevice=0;uDevice<cDevices;++uDevice) {
    SendMessageW(hWndList,LB_ADDSTRING,0,
        (LPARAM)pChannel[uDevice].vName.pwszVal);
  }

  SendMessageW(hWndList,LB_SETCURSEL,pPriv->uDevice,0);
list:
  return;
}

static void DevicesOnSelChange(HWND hDlg, Priv *pPriv, HWND hWndList)
{
  pPriv->uDevice=SendMessageW(hWndList,LB_GETCURSEL,0,0);
  ConfigSet(pPriv,pPriv->hDlg);
}

static INT_PTR CALLBACK DevicesProc(HWND hDlg, UINT uMsg, WPARAM wParam,
    LPARAM lParam)
{
  Priv *pPriv=(Priv *)GetWindowLongPtr(hDlg,GWLP_USERDATA);

  switch (uMsg) {
  case WM_INITDIALOG:
    pPriv=(Priv *)lParam;
    SetWindowLongPtr(hDlg,GWLP_USERDATA,lParam);
    DevicesFillList(hDlg,pPriv,IDC_LIST_DEVICES);
    return TRUE;
  case WM_COMMAND:
    switch (LOWORD(wParam)) {
    case IDC_LIST_DEVICES:
      switch (HIWORD(wParam)) {
      case LBN_SELCHANGE:
        DevicesOnSelChange(hDlg,pPriv,(HWND)lParam);
        return TRUE;
      default:
        break;
      }

      break;
    default:
      break;
    }

    break;
  default:
    break;
  }

  return FALSE;
}

static void DevicesSet(Page *pPage, Priv *pPriv)
{
}

static void DevicesGet(Page *pPage, Priv *pPriv)
{
  UINT *puDevice=&pPriv->pPlayer->uDevice;
  HWND hWndList;

  hWndList=GetDlgItem(pPage->hDlg,IDC_LIST_DEVICES);
  SendMessageW(hWndList,LB_GETSELITEMS,1,(LPARAM)puDevice);
}

static void DevicesReset(Page *pPage, Priv *pPriv)
{
}

static PageVMT *GetDevicesVMT(void)
{
  static PageVMT vmt;

  if (NULL==vmt.name) {
    vmt.name="Devices";
    vmt.lpDialogFunc=DevicesProc;
    vmt.Set=DevicesSet;
    vmt.Get=DevicesGet;
    vmt.Reset=DevicesReset;
  }

  return &vmt;
}

///////////////////////////////////////////////////////////////////////////////
static void OptionsSyncBalance(HWND hDlg)
{
  LRESULT bPush=!SendMessageW(GetDlgItem(hDlg,IDC_COMBOBOX_PULL),
      CB_GETCURSEL,0,0);
  LRESULT bBalance;

  EnableWindow(GetDlgItem(hDlg,IDC_CHECKBOX_BALANCE),bPush);

  if (bPush) {
    bBalance=SendMessageW(GetDlgItem(hDlg,IDC_CHECKBOX_BALANCE),
        BM_GETCHECK,0,0);
    EnableWindow(GetDlgItem(hDlg,IDC_SLIDER_BALANCE_START),bBalance);
    EnableWindow(GetDlgItem(hDlg,IDC_STATIC_BALANCE_START),bBalance);
  }
  else {
    EnableWindow(GetDlgItem(hDlg,IDC_SLIDER_BALANCE_START),FALSE);
    EnableWindow(GetDlgItem(hDlg,IDC_STATIC_BALANCE_START),FALSE);
  }
}

static void OptionsSyncShareMode(HWND hDlg)
{
  ChannelOptions options;

  ControlGet(gcaDeviceControlList,hDlg,&options);

  if (YASAPI_SHAREMODE_EXCLUSIVE==options.eShareMode) {
    EnableWindow(
      GetDlgItem(hDlg,IDC_CHECKBOX_AUTOCONVERT_PCM),
                                // _In_  HWND hWnd,
      FALSE                     // _In_  BOOL bEnable
    );

    EnableWindow(
      GetDlgItem(hDlg,IDC_CHECKBOX_SRC_DEFAULT_QUALITY),
                                // _In_  HWND hWnd,
      FALSE                     // _In_  BOOL bEnable
    );
  }
  else {
    EnableWindow(
      GetDlgItem(hDlg,IDC_CHECKBOX_AUTOCONVERT_PCM),
                                // _In_  HWND hWnd,
      TRUE                      // _In_  BOOL bEnable
    );

    EnableWindow(
      GetDlgItem(hDlg,IDC_CHECKBOX_SRC_DEFAULT_QUALITY),
                                // _In_  HWND hWnd,
      options.bAutoConvertPCM   // _In_  BOOL bEnable
    );
  }
}

static INT_PTR CALLBACK OptionsProc(HWND hDlg, UINT uMsg, WPARAM wParam,
    LPARAM lParam)
{
  Priv *pPriv=(Priv *)GetWindowLongPtr(hDlg,GWLP_USERDATA);

  switch (uMsg) {
  case WM_INITDIALOG:
    pPriv=(Priv *)lParam;
    SetWindowLongPtr(hDlg,GWLP_USERDATA,lParam);
    ControlInit(gcaDeviceControlList,hDlg,pPriv->hModule);
    return TRUE;
  case WM_COMMAND:
    switch (LOWORD(wParam)) {
    //case IDC_COMBOBOX_SHAREMODE:
    case IDC_CHECKBOX_AUTOCONVERT_PCM:
    case IDC_CHECKBOX_SRC_DEFAULT_QUALITY:
      OptionsSyncShareMode(hDlg);
      return TRUE;
    //case IDC_COMBOBOX_PULL:
    case IDC_CHECKBOX_BALANCE:
      OptionsSyncBalance(hDlg);
      return TRUE;
    case IDC_COMBOBOX_SHAREMODE:
      OptionsSyncShareMode(hDlg);
      SyncDevicePeriod(pPriv);
      return TRUE;
    case IDC_COMBOBOX_PULL:
      OptionsSyncBalance(hDlg);
      SyncDevicePeriod(pPriv);
      return TRUE;
    default:
      break;
    }

    break;
  case WM_HSCROLL:
    ControlSync(gcaDeviceControlList,hDlg);
    break;
  default:
    break;
  }

  return FALSE;
}

static void OptionsSetOptions(Page *pPage, Priv *pPriv,
    const ChannelOptions *pOptions)
{
  ControlSet(gcaDeviceControlList,pPage->hDlg,pOptions);
  OptionsSyncShareMode(pPage->hDlg);
  OptionsSyncBalance(pPage->hDlg);
  SyncDevicePeriod(pPriv);
}

static void OptionsSet(Page *pPage, Priv *pPriv)
{
  Player *pPlayer=pPriv->pPlayer;
  Channel *pChannel=pPlayer->pChannel+pPriv->uDevice;
  ChannelOptions *pOptions=&pChannel->options;

  SetWindowTextW(GetDlgItem(pPage->hDlg,IDC_STATIC_DEVICE_NAME),
      pChannel->vName.pwszVal);

  OptionsSetOptions(pPage,pPriv,pOptions);
}

static void OptionsGet(Page *pPage, Priv *pPriv)
{
  Player *pPlayer=pPriv->pPlayer;
  Channel *pChannel=pPlayer->pChannel+pPriv->uDevice;
  ChannelOptions *pOptions=&pChannel->options;

  ControlGet(gcaDeviceControlList,pPage->hDlg,pOptions);
}

static void OptionsReset(Page *pPage, Priv *pPriv)
{
  OptionsSetOptions(pPage,pPriv,ChannelOptionsDefault());
}

static PageVMT *GetOptionsVMT(void)
{
  static PageVMT vmt;

  if (NULL==vmt.name) {
    vmt.name="Options";
    vmt.lpDialogFunc=OptionsProc;
    vmt.Set=OptionsSet;
    vmt.Get=OptionsGet;
    vmt.Reset=OptionsReset;
  }

  return &vmt;
}

///////////////////////////////////////////////////////////////////////////////
static INT_PTR CALLBACK BuffersProc(HWND hDlg, UINT uMsg, WPARAM wParam,
    LPARAM lParam)
{
  Priv *pPriv=(Priv *)GetWindowLongPtr(hDlg,GWLP_USERDATA);

  switch (uMsg) {
  case WM_INITDIALOG:
    pPriv=(Priv *)lParam;
    SetWindowLongPtr(hDlg,GWLP_USERDATA,lParam);
    ControlInit(gcaBuffersControlList,hDlg,pPriv->hModule);
    return TRUE;
  case WM_HSCROLL:
    ControlSync(gcaBuffersControlList,hDlg);
    break;
  default:
    break;
  }

  return FALSE;
}

static void BuffersSet(Page *pPage, Priv *pPriv)
{
  Player *pPlayer=pPriv->pPlayer;
  Channel *pChannel=pPlayer->pChannel+pPriv->uDevice;
  ChannelOptions *pOptions=&pChannel->options;

  SetWindowTextW(GetDlgItem(pPage->hDlg,IDC_STATIC_DEVICE_NAME),
      pChannel->vName.pwszVal);
  ControlSet(gcaBuffersControlList,pPage->hDlg,pOptions);
}

static void BuffersGet(Page *pPage, Priv *pPriv)
{
  Player *pPlayer=pPriv->pPlayer;
  Channel *pChannel=pPlayer->pChannel+pPriv->uDevice;
  ChannelOptions *pOptions=&pChannel->options;

  ControlGet(gcaBuffersControlList,pPage->hDlg,pOptions);
}

static void BuffersReset(Page *pPage, Priv *pPriv)
{
  ControlSet(gcaBuffersControlList,pPage->hDlg,ChannelOptionsDefault());
}

static PageVMT *GetBuffersVMT(void)
{
  static PageVMT vmt;

  if (NULL==vmt.name) {
    vmt.name="Buffers";
    vmt.lpDialogFunc=BuffersProc;
    vmt.Set=BuffersSet;
    vmt.Get=BuffersGet;
    vmt.Reset=BuffersReset;
  }

  return &vmt;
}

///////////////////////////////////////////////////////////////////////////////
static void OnSelChange(HWND hDlg, Priv *pPriv, HWND hWndTab)
{
  int debug=pPriv->pPlayer->pPlugin->options.debug;
  int nPage=TabCtrl_GetCurSel(hWndTab);
  Page *pPage=pPriv->aPages+nPage;

  if (pPriv->hDlgPage) {
    ShowWindow(pPriv->hDlgPage,FALSE);
    pPriv->hDlgPage=NULL;
  }

  pPriv->nPage=nPage;
  pPriv->hDlgPage=pPage->hDlg;
  ShowWindow(pPriv->hDlgPage,TRUE);
}

static int InitPage(HWND hDlg, HWND hWndTab, Priv *pPriv, int nPage)
{
  int code=-1;
  PageTemplate *pTemplate=pPriv->pTemplates+nPage;
  Page *pPage=pPriv->aPages+nPage;
#if defined (YASAPI_DEBUG) // {
  int debug=pPriv->pPlayer->pPlugin->options.debug;
#endif // }
  TCITEMA tie;
  LPCTSTR lpTemplateName;
  RECT rc;

  pPage->vmt=pTemplate->GetVMT();

  memset(&tie,0,sizeof tie);
  tie.mask=TCIF_TEXT|TCIF_IMAGE;
  tie.iImage=-1;
  tie.pszText=pTemplate->label;
  SendMessageA(hWndTab,TCM_INSERTITEMA,nPage,(LPARAM)&tie);

  // Finding resource.
  lpTemplateName=MAKEINTRESOURCEA(pTemplate->idd);

  pPage->hDlg=CreateDialogParamA(
    pPriv->hModule,   // _In_opt_ HINSTANCE hInstance,
    lpTemplateName,   // _In_     LPCTSTR   lpTemplateName,
    hDlg,             // _In_opt_ HWND      hWndParent,
    pPage->vmt->lpDialogFunc,
                      // _In_opt_ DLGPROC   lpDialogFunc,
    (LPARAM)pPriv     // _In_     LPARAM    dwInitParam
  );

  if (NULL==pPage->hDlg) {
    DMESSAGE(debug,"creating child dialog");
    goto child;
  }

  GetClientRect(pPage->hDlg,&rc);
  TabCtrl_AdjustRect(hWndTab,FALSE,&rc);

  MapWindowPoints(
    hWndTab,          // _In_    HWND    hWndFrom,
    hDlg,             // _In_    HWND    hWndTo,
    (LPPOINT)&rc,     // _Inout_ LPPOINT lpPoints,
    2                 // _In_    UINT    cPoints
  );

  SetWindowPos(pPage->hDlg,HWND_TOP,rc.left,rc.top,0,0,
      SWP_NOSIZE|SWP_HIDEWINDOW);
  code=0;
child:
lock:
load:
find:
  return code;
}

static int InitTab(HWND hDlg, int idc, Priv *pPriv)
{
  int code=-1;
#if defined (YASAPI_DEBUG) // {
  int debug=pPriv->pPlayer->pPlugin->options.debug;
#endif // }
  HWND hWndTab;
  int nPage;

  // Get the tabcontrol handle.
  if (NULL==(hWndTab=GetDlgItem(hDlg,idc))) {
    DMESSAGE(debug,"getting tabcontrol handle");
    goto item;
  }

  // Initialize pages.
  for (nPage=0;nPage<pPriv->nNumPages;++nPage) {
    if (InitPage(hDlg,hWndTab,pPriv,nPage)<0) {
      DMESSAGE(debug,"initializing page");
      goto param;
    }

#if 1 // {
    EnableThemeDialogTexture(
      pPriv->aPages[nPage].hDlg,  // _In_ HWND  hwnd,
      ETDT_ENABLETAB              // _In_ DWORD dwFlags
    );
#endif // }
  }

  SendMessageA(hWndTab,TCM_SETCURSEL,pPriv->nPage,0);
  OnSelChange(hDlg,pPriv,hWndTab);
  code=0;
param:
item:
  return code;
}

#if defined (YASAPI_DEBUG) // {
static void SyncDebug(Priv *pPriv, HWND hDlg)
{
  Plugin *pPlugin=pPriv->pPlayer->pPlugin;
  LRESULT debug=SendMessageW(GetDlgItem(hDlg,IDC_COMBOBOX_DEBUG),
      CB_GETCURSEL,0,0);

  EnableWindow(GetDlgItem(hDlg,IDC_STATIC_LABEL_SLEEP),debug);
  EnableWindow(GetDlgItem(hDlg,IDC_COMBOBOX_SLEEP),debug);

  if (!pPlugin->options.debug&&debug)
    ConsoleOpen();
  else if (pPlugin->options.debug&&!debug)
    ConsoleClose();

  pPlugin->options.debug=debug;
}
#endif // }

static void ConfigSet(Priv *pPriv, HWND hDlg)
{
  int nPage;
  Page *pPage;

  ControlSet(gcaPluginControlList,hDlg,&pPriv->pPlayer->pPlugin->options);
#if defined (YASAPI_DEBUG) // {
  SyncDebug(pPriv,hDlg);
#endif // }

  for (nPage=0;nPage<pPriv->nNumPages;++nPage) {
    pPage=pPriv->aPages+nPage;
    pPage->vmt->Set(pPage,pPriv);
  }
}

static void ConfigGet(Priv *pPriv, HWND hDlg)
{
  Player *pPlayer=pPriv->pPlayer;
  Plugin *pPlugin=pPlayer->pPlugin;
#if defined (YASAPI_DEBUG) // {
  int debug=pPlugin->options.debug;
#endif // }
  Channel *pChannel;
  int nPage;
  Page *pPage;
  RECT rc;

  // critical section /////////////////////////////////////////////////////////
  QueueLockMutex(&pPlugin->queue); // {

  // get the coosen device/player.
  pPlayer->uDevice=pPriv->uDevice;
  pChannel=pPlayer->pChannel+pPlayer->uDevice;
  wcsncpy(pPlugin->options.id,pChannel->pstrId,YASAPI_ID_SIZE);
  // get the current tabcontrol page.
  pPlugin->options.page=pPriv->nPage;
  // read out the plugin options.
  ControlGet(gcaPluginControlList,hDlg,&pPlugin->options);

  // read out the device/player options.
  for (nPage=0;nPage<pPriv->nNumPages;++nPage) {
    pPage=pPriv->aPages+nPage;
    pPage->vmt->Get(pPage,pPriv);
  }

  QueueUnlockMutex(&pPlugin->queue); // }
  /////////////////////////////////////////////////////////////////////////////

  GetWindowRect(hDlg,&rc);
  pPlugin->options.posx=rc.left;
  pPlugin->options.posy=rc.top;

  // save the plugin options.
  PluginOptionsSave(&pPlugin->options);
  // save the device/player options.
  ChannelOptionsSave(&pChannel->options);

#if defined (YASAPI_DEBUG) // {
  if (debug!=pPlugin->options.debug) {
    if (pPlugin->options.debug)
      ConsoleOpen();
    else
      ConsoleClose();
  }
#endif // }
}

static void ConfigReset(Priv *pPriv, HWND hDlg)
{
  int nPage;
  Page *pPage;

  ControlSet(gcaPluginControlList,hDlg,PluginOptionsDefault());

  for (nPage=0;nPage<pPriv->nNumPages;++nPage) {
    pPage=pPriv->aPages+nPage;
    pPage->vmt->Reset(pPage,pPriv);
  }
}

static void ConfigEndDialog(HWND hDlg, INT_PTR nResult, Priv *pPriv)
{
  Plugin *pPlugin=pPriv->pPlayer->pPlugin;

  QueueLockMutex(&pPlugin->queue); // {
  pPlugin->hDlgConfig=NULL;
  QueueUnlockMutex(&pPlugin->queue); // }
  EndDialog(hDlg,nResult);
}

static void UpdateProgress(HWND hWnd, WORD wParam)
{
  LRESULT lMin=SendMessageW(hWnd,PBM_GETRANGE,TRUE,0);
  LRESULT lMax=SendMessageW(hWnd,PBM_GETRANGE,FALSE,0);
  WORD wPos=lMin+MulDiv(lMax-lMin,wParam,USHRT_MAX);

  SendMessageW(hWnd,PBM_SETPOS,wPos,0);
}

static void SaveConfigOptions(HWND hDlg, Priv *pPriv)
{
  PluginOptions *pOptions=&pPriv->pPlayer->pPlugin->options;
  const wchar_t *group=YASAPI_GROUP_GENERAL;
  const wchar_t *path=pOptions->path;
  RECT rc;

  GetWindowRect(hDlg,&rc);
  pOptions->posx=rc.left;
  pOptions->posy=rc.top;
  pOptions->page=pPriv->nPage;

  PropertySaveInt(group,YASAPI_PROPERTY_VIS,pOptions->bVisualization,path);
  PropertySaveInt(group,YASAPI_PROPERTY_POSX,pOptions->posx,path);
  PropertySaveInt(group,YASAPI_PROPERTY_POSY,pOptions->posy,path);
  PropertySaveInt(group,YASAPI_PROPERTY_PAGE,pOptions->page,path);
#if defined (YASAPI_DEBUG) // {
  PropertySaveInt(group,YASAPI_PROPERTY_DEBUG,pOptions->debug,path);
#endif // }
}

static void SyncVisualization(HWND hDlg, Priv *pPriv)
{
  Plugin *pPlugin=pPriv->pPlayer->pPlugin;
  BOOL bVisualization=IsDlgButtonChecked(hDlg,IDC_CHECKBOX_VISUALIZATION);
  HWND hWndRing=GetDlgItem(hDlg,IDC_PROGRESSBAR_RING);
  HWND hWndShared=GetDlgItem(hDlg,IDC_PROGRESSBAR_SHARED);

  if (!bVisualization) {
    UpdateProgress(hWndRing,0);
    UpdateProgress(hWndShared,0);
  }

  EnableWindow(GetDlgItem(hDlg,IDC_STATIC_RING),bVisualization);
  EnableWindow(hWndRing,bVisualization);
  EnableWindow(GetDlgItem(hDlg,IDC_STATIC_SHARED),bVisualization);
  EnableWindow(hWndShared,bVisualization);

  pPlugin->options.bVisualization=bVisualization;
}

static INT_PTR CALLBACK ConfigProc(HWND hDlg, UINT uMsg, WPARAM wParam,
    LPARAM lParam)
{
  Priv *pPriv=(Priv *)GetWindowLongPtr(hDlg,GWLP_USERDATA);
#if defined (YASAPI_DEBUG) // {
  int debug;
#endif // }

  switch (uMsg) {
  case WM_INITDIALOG:
    pPriv=(Priv *)lParam;
    pPriv->hDlg=hDlg;
#if defined (YASAPI_DEBUG) // {
    debug=pPriv->pPlayer->pPlugin->options.debug;
#endif // }
    SetWindowLongPtr(hDlg,GWLP_USERDATA,lParam);
    SetWindowTextA(hDlg,PI_LABEL " Settings");

    if (0<=pPriv->pPlayer->pPlugin->options.posx
        &&0<=pPriv->pPlayer->pPlugin->options.posy) {
      SetWindowPos(
        hDlg,   // _In_     HWND hWnd,
        NULL,   // _In_opt_ HWND hWndInsertAfter,
        pPriv->pPlayer->pPlugin->options.posx,
                // _In_     int  X,
        pPriv->pPlayer->pPlugin->options.posy,
                // _In_     int  Y,
        0,      // _In_     int  cx,
        0,      // _In_     int  cy,
        SWP_NOOWNERZORDER|SWP_NOSIZE
                //_In_     UINT uFlags
      );
    }

    ControlInit(gcaPluginControlList,hDlg,pPriv->hModule);

    if (InitTab(hDlg,IDC_TABCONTROL,pPriv)<0) {
      DMESSAGE(debug,"initializing tabcontrol");
      break;
    }

#if ! defined (YASAPI_DEBUG) // {
    EnableWindow(GetDlgItem(hDlg,IDC_STATIC_LABEL_DEBUG),FALSE);
    SendMessageW(GetDlgItem(hDlg,IDC_COMBOBOX_DEBUG),
        CB_ADDSTRING,0,(LPARAM)L"Off");
    SendMessageW(GetDlgItem(hDlg,IDC_COMBOBOX_DEBUG),
        CB_SETCURSEL,0,0);
    EnableWindow(GetDlgItem(hDlg,IDC_COMBOBOX_DEBUG),FALSE);

    EnableWindow(GetDlgItem(hDlg,IDC_STATIC_LABEL_SLEEP),FALSE);
    SendMessageW(GetDlgItem(hDlg,IDC_COMBOBOX_SLEEP),
        CB_ADDSTRING,0,(LPARAM)L"0 sec");
    SendMessageW(GetDlgItem(hDlg,IDC_COMBOBOX_SLEEP),
        CB_SETCURSEL,0,0);
    EnableWindow(GetDlgItem(hDlg,IDC_COMBOBOX_SLEEP),FALSE);
#endif // }
    ConfigSet(pPriv,hDlg);
    pPriv->pPlayer->pPlugin->hDlgConfig=hDlg;
    break;
  case WM_COMMAND:
    switch (LOWORD(wParam)) {
    case IDOK:
      ConfigGet(pPriv,hDlg);
      ConfigEndDialog(hDlg,TRUE,pPriv);
      return TRUE;
    case IDCANCEL:
      SaveConfigOptions(hDlg,pPriv);
      ConfigEndDialog(hDlg,FALSE,pPriv);
      return TRUE;
    case IDC_BUTTON_DEFAULT:
      ConfigReset(pPriv,hDlg);
      return TRUE;
    case IDC_CHECKBOX_VISUALIZATION:
      if (BN_CLICKED==HIWORD(wParam))
        SyncVisualization(hDlg,pPriv);

      return TRUE;
#if defined (YASAPI_DEBUG) // {
    case IDC_COMBOBOX_DEBUG:
      SyncDebug(pPriv,hDlg);
#endif // }
      return TRUE;
    default:
      break;
    }

    break;
  case WM_NOTIFY:
    switch (((NMHDR *)lParam)->idFrom) {
    case IDC_TABCONTROL:
      switch (((NMHDR *)lParam)->code) {
      case TCN_SELCHANGE:
        OnSelChange(hDlg,pPriv,((NMHDR *)lParam)->hwndFrom);
        return TRUE;
      default:
        break;
      }

      break;
    default:
      break;
    }

    break;
  case WM_CONFIG_UPDATE:
    if (IsDlgButtonChecked(hDlg,IDC_CHECKBOX_VISUALIZATION)) {
      UpdateProgress(GetDlgItem(hDlg,IDC_PROGRESSBAR_RING),LOWORD(lParam));
      UpdateProgress(GetDlgItem(hDlg,IDC_PROGRESSBAR_SHARED),HIWORD(lParam));
    }

    return TRUE;
  default:
    break;
  }

  return FALSE;
}

void ConfigDialog(HWND hWndParent, HINSTANCE hInstance, Player *pPlayer)
{
  Plugin *pPlugin=pPlayer->pPlugin;
#if defined (YASAPI_DEBUG) // {
  int debug=pPlugin->options.debug;
#endif // }
  int nNumPages,nPage;
  size_t size;
  Priv *pPriv;
  LPCWSTR lpTemplateName;

  nNumPages=(sizeof gaTemplates)/(sizeof *gaTemplates);
  size=(sizeof *pPriv)+nNumPages*(sizeof *pPriv->aPages);

  if (NULL==(pPriv=malloc(size))) {
    DMESSAGE(debug,"allocating priv");
    goto priv;
  }

  memset(pPriv,0,sizeof *pPriv);
  pPriv->pPlayer=pPlayer;
  pPriv->hModule=hInstance;
  pPriv->hDlg=NULL;
  pPriv->hDlgPage=NULL;
  pPriv->uDevice=pPlayer->uDevice;
  pPriv->nNumPages=nNumPages;
  pPriv->nPage=pPlugin->options.page<nNumPages?pPlugin->options.page:0;
  pPriv->pTemplates=gaTemplates;

  lpTemplateName=MAKEINTRESOURCEW(IDD1_CONFIG);

  DialogBoxParamW(
    pPriv->hModule,       // _In_opt_  HINSTANCE hInstance,
    lpTemplateName,       // _In_      LPCTSTR lpTemplateName,
    hWndParent,           // _In_opt_  HWND hWndParent,
    ConfigProc,           // _In_opt_  DLGPROC lpDialogFunc,
    (LPARAM)pPriv         // _In_      LPARAM dwInitParam
  );

  free(pPriv);
priv:
  return;
}
#endif // }
