/*
 * yasapi_payer.c
 * Copyright (C) 2015 Peter Belkner <pbelkner@snafu.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <yasapi.h>

int PlayerCreate(Player *pPlayer, Plugin *pPlugin)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlugin->options.debug;
#endif // }
  IMMDeviceEnumerator *pEnumerator;
  IMMDevice *pDevice;
  LPWSTR pstrId;
  IMMDeviceCollection *pCollection;
  UINT cDevices,uDevice;
  HRESULT hr;

  /////////////////////////////////////////////////////////////////////////////
  pPlayer->pPlugin=pPlugin;

  // create a device enumerator ///////////////////////////////////////////////
  hr=CoCreateInstance(
    &CLSID_MMDeviceEnumerator,  // _In_   REFCLSID rclsid,
    NULL,                       // _In_   LPUNKNOWN pUnkOuter,
    CLSCTX_ALL,                 // _In_   DWORD dwClsContext,
    &IID_IMMDeviceEnumerator,   // _In_   REFIID riid,
    (void**)&pEnumerator        // _Out_  LPVOID *ppv
  );

  if (FAILED(hr)) {
    DERROR(debug,REGDB_E_CLASSNOTREG,hr);
    DERROR(debug,CLASS_E_NOAGGREGATION,hr);
    DERROR(debug,E_NOINTERFACE,hr);
    DERROR(debug,E_POINTER,hr);
    DMESSAGE(debug,"creating device enumerator");
    goto enumerator;
  }

  pPlayer->pEnumerator=pEnumerator;
  DPUTS(debug,0,"  device enumerator created\n");

  // get the default audio device /////////////////////////////////////////////
  hr=pEnumerator->lpVtbl->GetDefaultAudioEndpoint(pEnumerator,
    eRender,                  // [in]   EDataFlow dataFlow,
    eMultimedia,              // [in]   ERole role,
    &pDevice                  // [out]  IMMDevice **ppDevice
  );

  if (FAILED(hr)) {
    DERROR(debug,E_POINTER,hr);
    DERROR(debug,E_INVALIDARG,hr);
    DERROR(debug,E_NOTFOUND,hr);
    DERROR(debug,E_OUTOFMEMORY,hr);
    DMESSAGE(debug,"getting default audio device");
    goto device;
  }

  // get the default device's id //////////////////////////////////////////////
  hr=pDevice->lpVtbl->GetId(pDevice,
    &pstrId                   // [out] LPWSTR *ppstrId
  );

  if (FAILED(hr)) {
    DERROR(debug,E_OUTOFMEMORY,hr);
    DERROR(debug,E_POINTER,hr);
    DMESSAGE(debug,"getting player device id");
    goto id;
  }

  // get a device collection //////////////////////////////////////////////////
  hr=pEnumerator->lpVtbl->EnumAudioEndpoints(pEnumerator,
    eRender,                  // [in]  EDataFlow           dataFlow,
    DEVICE_STATE_ACTIVE,      // [in]  DWORD               dwStateMask,
    &pCollection              // [out] IMMDeviceCollection **ppDevices
  );

  if (FAILED(hr)) {
    DERROR(debug,E_POINTER,hr);
    DERROR(debug,E_INVALIDARG,hr);
    DERROR(debug,E_OUTOFMEMORY,hr);
    DMESSAGE(debug,"getting device collection");
    goto collection;
  }

  pPlayer->pCollection=pCollection;
  DPUTS(debug,0,"  device collection created\n");

  // retrieving the count of the devices //////////////////////////////////////
  hr=pCollection->lpVtbl->GetCount(pCollection,
    &cDevices                 // [out] UINT *pcDevices
  );

  if (FAILED(hr)) {
    DERROR(debug,E_POINTER,hr);
    DMESSAGE(debug,"getting device count");
    goto count;
  }
  else if (0==cDevices) {
    DMESSAGE(debug,"finding devices");
    goto count;
  }

  pPlayer->cDevices=cDevices;
  DPRINTF(debug,0,"  got device count: %d\n",cDevices);

  /////////////////////////////////////////////////////////////////////////////
  if (NULL==(pPlayer->pChannel=calloc(cDevices,sizeof *pPlayer->pChannel))) {
    DMESSAGE(debug,"allocating player array");
    goto array;
  }

  DPUTS(debug,0,"  player array allocated\n");

  /////////////////////////////////////////////////////////////////////////////
  uDevice=0;
  pPlayer->uDevice=cDevices;

  while (uDevice<cDevices) {
    if (ChannelCreate(pPlayer->pChannel+uDevice,pPlayer,uDevice)<0) {
      DMESSAGE(debug,"creating player");
      goto player;
    }

    DPRINTF(debug,0,"  player %d created\n",uDevice);

    if (0==wcscmp(pPlugin->options.id,pPlayer->pChannel[uDevice].pstrId))
      pPlayer->uDevice=uDevice;
    else if (cDevices==pPlayer->uDevice) {
      if (0==wcscmp(pstrId,pPlayer->pChannel[uDevice].pstrId))
        pPlayer->uDevice=uDevice;
    }

    ++uDevice;
  }

  if (cDevices==pPlayer->uDevice) {
    DMESSAGE(debug,"finding device");
    goto find;
  }

  DWPRINTF(debug,0,L"  device: \"%s\"\n",
      pPlayer->pChannel[pPlayer->uDevice].vName.pwszVal);

  // success //////////////////////////////////////////////////////////////////
  CoTaskMemFree(pstrId);
  pDevice->lpVtbl->Release(pDevice);
  // link the plugin to this player (enable configuration dialog).
  pPlugin->pPlayer=pPlayer;

  return 0;
//cleanup:
find:
  while (0<uDevice) {
    --uDevice;
    ChannelDestroy(pPlayer->pChannel+uDevice);
  }
player:
  free(pPlayer->pChannel);
array:
count:
  pCollection->lpVtbl->Release(pCollection);
collection:
  CoTaskMemFree(pstrId);
id:
  pDevice->lpVtbl->Release(pDevice);
device:
  pEnumerator->lpVtbl->Release(pEnumerator);
enumerator:
  return -1;
}

void PlayerDestroy(Player *pPlayer)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->pPlugin->options.debug;
#endif // }
  IMMDeviceCollection *pCollection=pPlayer->pCollection;
  UINT uDevice;
  IMMDeviceEnumerator *pEnumerator=pPlayer->pEnumerator;

  /////////////////////////////////////////////////////////////////////////////
  // disable configuration.
  pPlayer->pPlugin->pPlayer=NULL;

  /////////////////////////////////////////////////////////////////////////////
  uDevice=pPlayer->cDevices;

  while (0<uDevice) {
    --uDevice;
    DPRINTF(debug,0,"  destroying player %d\n",uDevice);
    ChannelDestroy(pPlayer->pChannel+uDevice);
  }

  DPUTS(debug,0,"  freeing player array\n");
  free(pPlayer->pChannel);

  DPUTS(debug,0,"  destroying device collection\n");
  pCollection->lpVtbl->Release(pCollection);

  DPUTS(debug,0,"  destroying device enumerator\n");
  pEnumerator->lpVtbl->Release(pEnumerator);

  /////////////////////////////////////////////////////////////////////////////
  pPlayer->pPlugin=NULL;
}

///////////////////////////////////////////////////////////////////////////////
HWND PlayerGetDlgConfig(Player *pPlayer)
{
  Plugin *pPlugin=pPlayer->pPlugin;

  return pPlugin->options.bVisualization?pPlugin->hDlgConfig:NULL;
}

void PlayerPostUpdate(Player *pPlayer, WORD wRing, WORD wShared)
{
  HWND hDlgConfig=PlayerGetDlgConfig(pPlayer);

  if (hDlgConfig)
    PostMessage(hDlgConfig,WM_CONFIG_UPDATE,0,MAKELONG(wRing,wShared));
}
