/*
 * yasapi_plugin.c
 * Copyright (C) 2015 Peter Belkner <pbelkner@snafu.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <yasapi.h>

///////////////////////////////////////////////////////////////////////////////
int PluginCreate(Plugin *pPlugin, const wchar_t *path, int debug)
{
  Queue *pQueue=&pPlugin->queue;
  Store *pStore=&pPlugin->store;
  Result *pResult;
  Request *pRequest;
  int state;
  HRESULT hr;

  /////////////////////////////////////////////////////////////////////////////
  pPlugin->qVolume=1.0;
  pPlugin->pPlayer=NULL;
  pPlugin->hDlgConfig=NULL;

  // create the plugin options ////////////////////////////////////////////////
  if (PluginOptionsCreate(&pPlugin->options,path,debug)<0) {
    DMESSAGE(debug,"creating plugin options");
    goto options;
  }

  DPUTS(debug,0,"  plugin options created\n");

  /////////////////////////////////////////////////////////////////////////////
  if (QueueCreate(&pPlugin->queue,debug)<0) {
    DMESSAGE(debug,"creating queue");
    goto queue;
  }

  DPUTS(debug,0,"  queue created\n");

  /////////////////////////////////////////////////////////////////////////////
  if (StoreCreate(&pPlugin->store,debug)<0) {
    DMESSAGE(debug,"creating store");
    goto store;
  }

  DPUTS(debug,0,"  store created\n");

  /////////////////////////////////////////////////////////////////////////////
  pPlugin->hEvent=CreateEvent(
    0,        // _In_opt_ LPSECURITY_ATTRIBUTES lpEventAttributes,
    FALSE,    // _In_     BOOL                  bManualReset,
    FALSE,    // _In_     BOOL                  bInitialState,
    NULL      // _In_opt_ LPCTSTR               lpName
  );

  if (NULL==pPlugin->hEvent) {
    DMESSAGE(debug,"creating event");
    goto event;
  }
  
  DPUTS(debug,0,"  event created\n");

  // create the timer /////////////////////////////////////////////////////////
  pPlugin->hTimer=CreateWaitableTimer(
    NULL,       // _In_opt_  LPSECURITY_ATTRIBUTES lpTimerAttributes,
    FALSE,      // _In_      BOOL bManualReset,
    NULL        // _In_opt_  LPCTSTR lpTimerName
  );

  if (NULL==pPlugin->hTimer) {
    DMESSAGE(debug,"creating timer");
    goto timer;
  }

  DPUTS(debug,0,"  waitable timer created\n");

  // create the thread ////////////////////////////////////////////////////////
  pPlugin->hThread=CreateThread(
    NULL,           // _In_opt_   LPSECURITY_ATTRIBUTES lpThreadAttributes,
    0,              // _In_       SIZE_T dwStackSize,
    PluginThread,   // _In_       LPTHREAD_START_ROUTINE lpStartAddress,
    pPlugin,        // _In_opt_   LPVOID lpParameter,
    0,              // _In_       DWORD dwCreationFlags,
    NULL            // _Out_opt_  LPDWORD lpThreadId
  );

  if (NULL==pPlugin->hThread) {
    DMESSAGE(debug,"creating thread");
    goto thread;
  }

  DPUTS(debug,0,"  thread created\n");

  // send the start message ///////////////////////////////////////////////////
  pResult=StoreGet(pStore);
  pRequest=QueueLockWrite(pQueue,pResult,REQUEST_CREATE);
  QueueUnlockWrite(pQueue,pResult);
    state=pResult->state;
  StorePut(pStore,pResult);

  if (state<0) {
    DMESSAGE(debug,"starting loop");
    goto failed;
  }

  DPUTS(debug,0,"  loop started\n");

  return 0;
//cleanup:
failed:
  QueueWrite(&pPlugin->queue,REQUEST_STOP,&pPlugin->store);
  WaitForSingleObject(pPlugin->hThread,INFINITE);
  CloseHandle(pPlugin->hThread);
thread:
  CloseHandle(pPlugin->hTimer);
timer:
  CloseHandle(pPlugin->hEvent);
event:
  StoreDestroy(&pPlugin->store,debug);
store:
  QueueDestroy(&pPlugin->queue,debug);
queue:
  PluginOptionsDestroy(&pPlugin->options);
options:
  return -1;
}

void PluginDestroy(Plugin *pPlugin)
{
  int debug=pPlugin->options.debug;

  DPUTS(debug,0,"  waiting for loop to end\n");
  QueueWrite(&pPlugin->queue,REQUEST_DESTROY,&pPlugin->store);

  DPUTS(debug,0,"  waiting for thread to return\n");
  WaitForSingleObject(pPlugin->hThread,INFINITE);

  DPUTS(debug,0,"  destroying thread\n");
  CloseHandle(pPlugin->hThread);

  DPUTS(debug,0,"  destroying timer\n");
  CloseHandle(pPlugin->hTimer);

  DPUTS(debug,0,"  destroying event\n");
  CloseHandle(pPlugin->hEvent);

  DPUTS(debug,0,"  destroying store\n");
  StoreDestroy(&pPlugin->store,debug);

  DPUTS(debug,0,"  destroying queue\n");
  QueueDestroy(&pPlugin->queue,debug);

  DPUTS(debug,0,"  destroying plugin options\n");
  PluginOptionsDestroy(&pPlugin->options);
}

///////////////////////////////////////////////////////////////////////////////
#if defined (YASAPI_DEBUG) // {
static void PluginPrintStats(Plugin *pPlugin)
{
  int debug=pPlugin->options.debug;

  DPRINTF(debug,0,"    min free reqests: %d (%d)\n",
      pPlugin->queue.nMinAvailable,QUEUE_SIZE);
  DPRINTF(debug,0,"    min free results: %d (%d)\n",
      pPlugin->store.nMinAvailable,STORE_SIZE);
}
#endif // }

static HANDLE PluginGetEvent(Plugin *pPlugin, Session *pSession)
{
   return !pSession->state.valid
      ||pSession->pStrategy->IsPush(pSession)
      ||pSession->state.flush
      ||!SESSION_IS_RUNNING(pSession)
      ?NULL
      :pPlugin->hEvent;
}

static void PluginRelease(Result **ppResultPause, Result **ppResultWrite)
{
  if (*ppResultPause) {
    (*ppResultPause)->state=0;
    SetEvent((*ppResultPause)->hEvent);
    *ppResultPause=NULL;
  }

  if (*ppResultWrite) {
    (*ppResultWrite)->state=0;
    SetEvent((*ppResultWrite)->hEvent);
    *ppResultWrite=NULL;
  }
}

DWORD WINAPI PluginThread(LPVOID lpParameter)
{
  Plugin *pPlugin=lpParameter;
  int *pnDebug=&pPlugin->options.debug;
  Queue *pQueue=&pPlugin->queue;
  DWORD dwTaskIndex=0;
  Result *pResultPause=NULL;
  Result *pResultWrite=NULL;
  SessionWriteConfig write;
  HANDLE hTask;
  HANDLE hEvent;
  Request *pRequest;
  Result *pResult;
  Player player;
  Session session;

  hTask=AvSetMmThreadCharacteristicsW(L"Pro Audio",&dwTaskIndex);
  ZeroMemory(&player,sizeof player);
  ZeroMemory(&session,sizeof session);

  for (;;) {
    hEvent=PluginGetEvent(pPlugin,&session);
    pRequest=QueueLockRead(pQueue,hEvent,SessionSendRead,&session);
    pResult=pRequest->pResult;

    switch (pRequest->tag) {
    case REQUEST_STOP:
      DPUTS(*pnDebug,0,"  REQUEST_STOP\n");
      QueueUnlockRead(&pPlugin->queue,pResult);
      goto end;
    case REQUEST_CREATE:
      DPUTS(*pnDebug,0,"  REQUEST_CREATE\n");
      pResult->state=PlayerCreate(&player,pPlugin);
      QueueUnlockRead(&pPlugin->queue,pResult);
      break;
    case REQUEST_DESTROY:
      DPUTS(*pnDebug,0,"  REQUEST_DESTROY\n");
#if defined (YASAPI_DEBUG) // {
      PluginPrintStats(pPlugin);
#endif // }
      PlayerDestroy(&player);
      QueueWrite(pQueue,REQUEST_STOP,NULL);
      QueueUnlockRead(&pPlugin->queue,pResult);
      break;
    case REQUEST_OPEN:
      DPUTS(*pnDebug,0,"  REQUEST_OPEN\n");
      pResult->state=SessionCreate(&session,&pRequest->open,
          player.pChannel+player.uDevice);
      QueueUnlockRead(&pPlugin->queue,pResult);
      break;
    case REQUEST_CLOSE:
      DPUTS(*pnDebug,0,"  REQUEST_CLOSE\n");
      PluginRelease(&pResultPause,&pResultWrite);

      if (session.state.valid) {
#if defined (YASAPI_DEBUG) // {
        PluginPrintStats(pPlugin);
#endif // }
        SessionStop(&session,TRUE);
        PlayerPostUpdate(&player,0,0);
        SessionDestroy(&session);
      }

      QueueUnlockRead(&pPlugin->queue,pResult);
      break;
    case REQUEST_WRITE:
      DPUTS(*pnDebug,2,"  REQUEST_WRITE\n");

      if (!session.state.valid)
        pResult->state=0;
      else if (!pPlugin->options.bWriteBlock)
        pResult->state=SessionWrite(&session,&pRequest->write);
      else if (SessionWriteNext(&session,&pRequest->write)) {
        // defer returning to the "write" client.
        write=pRequest->write;
        pResultWrite=pResult;
        pResult=NULL;
      }
      else
        pResult->state=0;

      QueueUnlockRead(&pPlugin->queue,pResult);
      break;
    case REQUEST_READ:
      DPUTS(*pnDebug,2,"  REQUEST_READ\n");

      if (session.state.flush)
        session.state.eof=1;
      else {
        SessionRead(&session);

        if (pResultWrite&&!SessionWriteNext(&session,&write)) {
          if (pResultPause) {
            // don't stop before the last cached package is completely
            // written to the ring buffer: that's now.
            SessionStop(&session,FALSE);
            pResultPause->state=SessionGetTime(&session);
            session.state.pause=1;
            SetEvent(pResultPause->hEvent);
            pResultPause=NULL;
          }

          pResultWrite->state=0;
          SetEvent(pResultWrite->hEvent);
          pResultWrite=NULL;
        }
      }

      QueueUnlockRead(&pPlugin->queue,pResult);
      break;
    case REQUEST_PAUSE:
      DPUTS(*pnDebug,0,"  REQUEST_PAUSE\n");

      if (session.state.valid) {
        // don't overwrite "session.state.pause" before start/stop!
        if (!pRequest->pause.state) {
          SessionStart(&session);
          session.state.pause=0;
          pResult->state=SessionGetTime(&session);
        }
        else if (!pResultWrite) {
          SessionStop(&session,FALSE);
          session.state.pause=1;
          pResult->state=SessionGetTime(&session);
        } 
        else {
          // we've a cached "write" request. don't stop before the last
          // cached package is completely written to the ring buffer.
          pResultPause=pResult;
          pResult=NULL;
        }
      }
      else
        pResult->state=0;

      QueueUnlockRead(&pPlugin->queue,pResult);
      break;
    case REQUEST_FLUSH:
      DPUTS(*pnDebug,0,"  REQUEST_FLUSH\n");

      if (session.state.valid) {
        PluginRelease(&pResultPause,&pResultWrite);
        RingReset(&session.ring);

        if (session.state.eof) {
          SessionStop(&session,TRUE);
          session.state.eof=0;
          session.state.flush=0;
          session.state.drain=0;
        }

        pResult->state=SessionGetTime(&session);
      }
      else
        pResult->state=0;

      QueueUnlockRead(&pPlugin->queue,pResult);
      break;
    case REQUEST_CANWRITE:
      DPUTS(*pnDebug,2,"  REQUEST_CANWRITE\n");
      pResult->state=session.state.valid?SessionCanWrite(&session):0;
      QueueUnlockRead(&pPlugin->queue,pResult);
      break;
    case REQUEST_ISPLAYING:
      DPUTS(*pnDebug,0,"  REQUEST_ISPLAYING\n");

      if (session.state.valid) {
        // we're draining.
        session.state.drain=1;
        // report back wether we're eof.
        pResult->state=!session.state.eof;
      }
      else
        pResult->state=0;

      QueueUnlockRead(&pPlugin->queue,pResult);
      break;
    case REQUEST_GETTIME:
      DPUTS(*pnDebug,3,"  REQUEST_GETTIME\n");
      pResult->state=session.state.valid?SessionGetTime(&session):0;
      QueueUnlockRead(&pPlugin->queue,pResult);
      break;
    default:
      DPRINTF(*pnDebug,0,"  WARNING: unknown request: %d\n",pRequest->tag);
      QueueUnlockRead(&pPlugin->queue,pResult);
      break;
    }
  }
end:
  if (hTask)
    AvRevertMmThreadCharacteristics(hTask);

  return 0;
}
