/*
 * yasapi_strategy.c
 * Copyright (C) 2015 Peter Belkner <pbelkner@snafu.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <yasapi.h>

///////////////////////////////////////////////////////////////////////////////
static HRESULT InitializePush(Session *pSession, AUDCLNT_SHAREMODE eShareMode,
    DWORD dwStreamFlags, REFERENCE_TIME hnsDevicePeriod)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pSession->pChannel->pPlayer->pPlugin->options.debug;
#endif // }
  IAudioClient *pClient=pSession->pClient;
  WAVEFORMATEX *pwfx=&pSession->wfxx.Format;

  DPRINTF(debug,0,"  intitializing audio client %s/PUSH with %I64d/0 hns\n",
      YASAPI_SHAREMODE_LABEL(eShareMode),hnsDevicePeriod);

  return pClient->lpVtbl->Initialize(pClient,
    eShareMode,           // [in]  AUDCLNT_SHAREMODE ShareMode,
    dwStreamFlags,        // [in]  DWORD StreamFlags,
    hnsDevicePeriod,      // [in]  REFERENCE_TIME hnsBufferDuration,
    0,                    // [in]  REFERENCE_TIME hnsPeriodicity,
    pwfx,                 // [in]  const WAVEFORMATEX *pFormat,
    NULL                  // [in]  LPCGUID AudioSessionGuid
  );
}

static void BalanceStartPush(Session *pSession)
{
  Channel *pChannel=pSession->pChannel;
  int bBalance=pChannel->options.bBalance;

  pSession->balance.q=bBalance?pChannel->options.qBalanceStart:0.5;
}

static void BalanceIteratePush(Session *pSession, UINT32 uFramesWrite)
{
  Channel *pChannel=pSession->pChannel;
  int bBalance=pChannel->options.bBalance;
  double q;

  if (bBalance&&0ull<uFramesWrite) {
    q=0.5*pSession->balance.q*pSession->uFrames/uFramesWrite;
    pSession->balance.q
        +=(q-pSession->balance.q)/(++pSession->balance.count);
  }
}

static void SetTimerPush(Session *pSession, LONGLONG time, BOOL bFlush)
{
  SessionSetTimer(pSession,time?pSession->balance.q*time+0.5:0);
}

static BOOL NeedDevicePeriodPush(Session *pSession,
    AUDCLNT_SHAREMODE eShareMode)
{
  return TRUE;
}

static BOOL CanAdaptPush(Session *pSession)
{
  return TRUE;
}

static BOOL IsPushPush(Session *pSession)
{
  return TRUE;
}

static UINT32 GetFramesPaddingPush(Session *pSession)
{
  return SessionGetFramesPadding(pSession);
}

const Strategy gcsPush={
  InitializePush,
  BalanceStartPush,
  BalanceIteratePush,
  SetTimerPush,
  NeedDevicePeriodPush,
  CanAdaptPush,
  IsPushPush,
  GetFramesPaddingPush
};

///////////////////////////////////////////////////////////////////////////////
static HRESULT InitializePull(Session *pSession, AUDCLNT_SHAREMODE eShareMode,
    DWORD dwStreamFlags, REFERENCE_TIME hnsDevicePeriod)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pSession->pChannel->pPlayer->pPlugin->options.debug;
#endif // }
  IAudioClient *pClient=pSession->pClient;
  WAVEFORMATEX *pwfx=&pSession->wfxx.Format;

  dwStreamFlags|=AUDCLNT_STREAMFLAGS_EVENTCALLBACK;

  switch (eShareMode) {
  case AUDCLNT_SHAREMODE_SHARED:
    // For a shared-mode stream that uses event-driven buffering, the
    // caller must set both hnsPeriodicity and hnsBufferDuration to 0. The
    // Initialize method determines how large a buffer to allocate based on
    // the scheduling period of the audio engine. Although the client's buffer
    // processing thread is event driven, the basic buffer management process,
    // as described previously, is unaltered. Each time the thread awakens,
    // it should call IAudioClient::GetCurrentPadding to determine how much
    // data to write to a rendering buffer or read from a capture buffer. In
    // contrast to the two buffers that the Initialize method allocates for an
    // exclusive-mode stream that uses event-driven buffering, a shared-mode
    // stream requires a single buffer.
    DPUTS(debug,0,"  intitializing audio client SHARED/PULL with 0/0 hns\n");

    return pClient->lpVtbl->Initialize(pClient,
      eShareMode,         // [in]  AUDCLNT_SHAREMODE ShareMode,
      dwStreamFlags,      // [in]  DWORD StreamFlags,
      0,                  // [in]  REFERENCE_TIME hnsBufferDuration,
      0,                  // [in]  REFERENCE_TIME hnsPeriodicity,
      pwfx,               // [in]  const WAVEFORMATEX *pFormat,
      NULL                // [in]  LPCGUID AudioSessionGuid
    );
  case AUDCLNT_SHAREMODE_EXCLUSIVE:
    // For an exclusive-mode stream that uses event-driven buffering,
    // the caller must specify nonzero values for hnsPeriodicity
    // and hnsBufferDuration, and the values of these two parameters must
    // be equal. The Initialize method allocates two buffers for the stream.
    // Each buffer is equal in duration to the value of the hnsBufferDuration
    // parameter. Following the Initialize call for a rendering stream, the
    // caller should fill the first of the two buffers before starting the
    // stream. For a capture stream, the buffers are initially empty, and
    // the caller should assume that each buffer remains empty until the
    // event for that buffer is signaled. While the stream is running, the
    // system alternately sends one buffer or the other to the client—this
    // form of double buffering is referred to as "ping-ponging". Each time
    // the client receives a buffer from the system (which the system
    // indicates by signaling the event), the client must process the entire
    // buffer. For example, if the client requests a packet size from the
    // IAudioRenderClient::GetBuffer method that does not match the buffer
    // size, the method fails. Calls to the IAudioClient::GetCurrentPadding
    // method are unnecessary because the packet size must always equal the
    // buffer size. In contrast to the buffering modes discussed previously,
    // the latency for an event-driven, exclusive-mode stream depends directly
    // on the buffer size.
    DPRINTF(debug,0,"  intitializing audio client EXCLUSIVE/PULL"
        " with %I64d/%I64d hns\n",hnsDevicePeriod,hnsDevicePeriod);

    return pClient->lpVtbl->Initialize(pClient,
      eShareMode,         // [in]  AUDCLNT_SHAREMODE ShareMode,
      dwStreamFlags,      // [in]  DWORD StreamFlags,
      hnsDevicePeriod,    // [in]  REFERENCE_TIME hnsBufferDuration,
      hnsDevicePeriod,    // [in]  REFERENCE_TIME hnsPeriodicity,
      pwfx,               // [in]  const WAVEFORMATEX *pFormat,
      NULL                // [in]  LPCGUID AudioSessionGuid
    );
  default:
    return E_FAIL;
  }
}

static void BalanceStartPull(Session *pSession)
{
  switch (pSession->eShareMode) {
  case AUDCLNT_SHAREMODE_SHARED:
    pSession->balance.q=0.5;
    break;
  case AUDCLNT_SHAREMODE_EXCLUSIVE:
    pSession->balance.q=1.0;
    break;
  default:
    break;
  }
}

static void BalanceIteratePull(Session *pSession, UINT32 uFramesWrite)
{
}

static void SetTimerPull(Session *pSession, LONGLONG time, BOOL bFlush)
{
  if (bFlush)
    SessionSetTimer(pSession,pSession->balance.q*time+0.5);
}

static BOOL NeedDevicePeriodPull(Session *pSession,
    AUDCLNT_SHAREMODE eShareMode)
{
  switch (eShareMode) {
  case AUDCLNT_SHAREMODE_SHARED:
    return FALSE;
  case AUDCLNT_SHAREMODE_EXCLUSIVE:
    return TRUE;
  default:
    return TRUE;
  }
}

static BOOL CanAdaptPull(Session *pSession)
{
  switch (pSession->eShareMode) {
  case AUDCLNT_SHAREMODE_SHARED:
    return TRUE;
  case AUDCLNT_SHAREMODE_EXCLUSIVE:
    return FALSE;
  default:
    return TRUE;
  }
}

static BOOL IsPushPull(Session *pSession)
{
  return FALSE;
}

static UINT32 GetFramesPaddingPull(Session *pSession)
{
  switch (pSession->eShareMode) {
  case AUDCLNT_SHAREMODE_SHARED:
    return SessionGetFramesPadding(pSession);
  case AUDCLNT_SHAREMODE_EXCLUSIVE:
    return 0;
  default:
    return SessionGetFramesPadding(pSession);
  }
}

const Strategy gcsPull={
  InitializePull,
  BalanceStartPull,
  BalanceIteratePull,
  SetTimerPull,
  NeedDevicePeriodPull,
  CanAdaptPull,
  IsPushPull,
  GetFramesPaddingPull
};
